@@ -1,7 +1,7 @@
 /*****************************************************************************/
 /*  LibreDWG - free implementation of the DWG file format                    */
 /*                                                                           */
-/*  Copyright (C) 2010, 2018 Free Software Foundation, Inc.                  */
+/*  Copyright (C) 2010-2019 Free Software Foundation, Inc.                   */
 /*                                                                           */
 /*  This library is free software, licensed under the terms of the GNU       */
 /*  General Public License as published by the Free Software Foundation,     */
@@ -16,12 +16,14 @@
  * modified by Reini Urban
  */
 
+#include "config.h"
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdint.h>
 #include <sys/types.h>
 #include <stdbool.h>
 #include <assert.h>
+#include "common.h"
 #include "bits.h"
 #include "dec_macros.h"
 #include "decode.h"
@@ -61,7 +63,7 @@ typedef struct r2007_file_header
   int64_t unknown3; // 0xf800
   int64_t unknown4; // 4
   int64_t unknown5; // 1
-  int64_t sections_amount;
+  int64_t num_sections;
   int64_t sections_map_crc_uncomp;
   int64_t sections_map_size_comp;
   int64_t sections_map2_id;
@@ -126,7 +128,7 @@ static void pages_destroy (r2007_page *page);
 static void sections_destroy (r2007_section *section);
 static r2007_section *read_sections_map (Bit_Chain *dat, int64_t size_comp,
                                          int64_t size_uncomp,
-                                         int64_t correction);
+                                         int64_t correction) ATTRIBUTE_MALLOC;
 static int read_data_section (Bit_Chain *sec_dat, Bit_Chain *dat,
                               r2007_section *restrict sections_map,
                               r2007_page *restrict pages_map,
@@ -143,8 +145,17 @@ static int read_2007_section_handles (Bit_Chain *dat, Bit_Chain *hdl_dat,
                                       Dwg_Data *restrict dwg,
                                       r2007_section *restrict sections_map,
                                       r2007_page *restrict pages_map);
+static int read_2007_section_summary (Bit_Chain *restrict dat,
+                                      Dwg_Data *restrict dwg,
+                                      r2007_section *restrict sections_map,
+                                      r2007_page *restrict pages_map);
+static int read_2007_section_preview (Bit_Chain *restrict dat,
+                                      Dwg_Data *restrict dwg,
+                                      r2007_section *restrict sections_map,
+                                      r2007_page *restrict pages_map);
 static r2007_page *read_pages_map (Bit_Chain *dat, int64_t size_comp,
-                                   int64_t size_uncomp, int64_t correction);
+                                   int64_t size_uncomp,
+                                   int64_t correction) ATTRIBUTE_MALLOC;
 static int read_file_header (Bit_Chain *restrict dat,
                              r2007_file_header *restrict file_header);
 static void read_instructions (BITCODE_RC *restrict *restrict src,
@@ -161,9 +172,9 @@ static uint32_t read_literal_length (BITCODE_RC *restrict *restrict src,
 static void copy_compressed_bytes (BITCODE_RC *restrict dst,
                                    BITCODE_RC *restrict src, int length);
 static DWGCHAR *bfr_read_string (BITCODE_RC *restrict *restrict src,
-                                 int64_t size);
+                                 int64_t size) ATTRIBUTE_MALLOC;
 static BITCODE_RC *decode_rs (const BITCODE_RC *src, int block_count,
-                              int data_size);
+                              int data_size) ATTRIBUTE_MALLOC;
 static int decompress_r2007 (BITCODE_RC *restrict dst, int dst_size,
                              BITCODE_RC *restrict src, int src_size);
 
@@ -185,8 +196,9 @@ static int decompress_r2007 (BITCODE_RC *restrict dst, int dst_size,
   memcpy (&dst[8], &src[offset], 8);                                          \
   dst += 16
 
+/* Don't use restrict here: GH #141 broken for most newer compilers */
 static inline BITCODE_RC *
-copy_bytes_2 (BITCODE_RC *restrict dst, const BITCODE_RC *restrict src)
+copy_bytes_2 (BITCODE_RC *dst, const BITCODE_RC *src)
 {
   dst[0] = src[1];
   dst[1] = src[0];
@@ -194,7 +206,7 @@ copy_bytes_2 (BITCODE_RC *restrict dst, const BITCODE_RC *restrict src)
 }
 
 static inline BITCODE_RC *
-copy_bytes_3 (BITCODE_RC *restrict dst, const BITCODE_RC *restrict src)
+copy_bytes_3 (BITCODE_RC *dst, const BITCODE_RC *src)
 {
   dst[0] = src[2];
   dst[1] = src[1];
@@ -467,6 +479,7 @@ decompress_r2007 (BITCODE_RC *restrict dst, int dst_size,
   uint32_t length = 0;
   uint32_t offset = 0;
 
+  BITCODE_RC *dst_start = dst;
   BITCODE_RC *dst_end = dst + dst_size;
   BITCODE_RC *src_end = src + src_size;
   unsigned char opcode;
@@ -524,7 +537,12 @@ decompress_r2007 (BITCODE_RC *restrict dst, int dst_size,
               LOG_ERROR ("Decompression error: length overflow");
               return DWG_ERR_INTERNALERROR;
             }
-          // LOG_INSANE("copy_bytes(%p %u %u)\n", dst, length, offset);
+          if (offset > (uint32_t) (dst - dst_start))
+            {
+              LOG_ERROR ("Decompression error: offset underflow");
+              return DWG_ERR_INTERNALERROR;
+            }
+          LOG_INSANE ("copy_bytes(%p %u %u)\n", dst, length, offset);
           copy_bytes (dst, length, offset);
 
           dst += length;
@@ -624,13 +642,23 @@ read_system_page (Bit_Chain *dat, int64_t size_comp, int64_t size_uncomp,
   rsdata = &data[size_uncomp];
   bit_read_fixed (dat, rsdata, page_size);
   pedata = decode_rs (rsdata, block_count, 239);
+  if (!pedata)
+    {
+      free (data);
+      return NULL;
+    }
 
   if (size_comp < size_uncomp)
-    (void)decompress_r2007 (data, size_uncomp, pedata, size_comp);
+    error = decompress_r2007 (data, size_uncomp, pedata, size_comp);
   else
     memcpy (data, pedata, size_uncomp);
   free (pedata);
 
+  if (error >= DWG_ERR_CRITICAL)
+    {
+      free (data);
+      return NULL;
+    }
   return data;
 }
 
@@ -659,6 +687,8 @@ read_data_page (Bit_Chain *restrict dat, BITCODE_RC *restrict decomp,
     }
   bit_read_fixed (dat, rsdata, page_size);
   pedata = decode_rs (rsdata, block_count, 0xFB);
+  if (!pedata)
+    return DWG_ERR_OUTOFMEM;
 
   if (size_comp < size_uncomp)
     error = decompress_r2007 (decomp, size_uncomp, pedata, size_comp);
@@ -680,7 +710,7 @@ read_data_section (Bit_Chain *sec_dat, Bit_Chain *dat,
   r2007_page *page;
   uint64_t max_decomp_size;
   BITCODE_RC *decomp;
-  int error, i;
+  int error = 0, i;
 
   section = get_section (sections_map, sec_type);
   if (section == NULL)
@@ -697,6 +727,11 @@ read_data_section (Bit_Chain *sec_dat, Bit_Chain *dat,
       return DWG_ERR_OUTOFMEM;
     }
 
+  sec_dat->bit = 0;
+  sec_dat->byte = 0;
+  sec_dat->size = max_decomp_size;
+  sec_dat->version = dat->version;
+
   for (i = 0; i < (int)section->num_pages; i++)
     {
       r2007_section_page *section_page = section->pages[i];
@@ -716,23 +751,27 @@ read_data_section (Bit_Chain *sec_dat, Bit_Chain *dat,
         }
 
       dat->byte = page->offset;
-      error = read_data_page (dat, &decomp[section_page->offset], page->size,
-                              section_page->comp_size,
-                              section_page->uncomp_size);
-      if (error)
+      // only if compressed. TODO: Isn't there a compressed flag as with 2004+?
+      // theoretically the sizes could still be the same.
+      if (section_page->comp_size != section_page->uncomp_size)
         {
-          free (decomp);
-          LOG_ERROR ("Failed to read page")
-          return error;
+          error = read_data_page (dat, &decomp[section_page->offset],
+                                  page->size, section_page->comp_size,
+                                  section_page->uncomp_size);
+          if (error)
+            {
+              free (decomp);
+              LOG_ERROR ("Failed to read compressed page")
+              return error;
+            }
+        }
+      else
+        {
+          memcpy (&decomp[section_page->offset], &dat->chain[dat->byte],
+                  section_page->uncomp_size);
         }
     }
-
-  sec_dat->bit = 0;
-  sec_dat->byte = 0;
   sec_dat->chain = decomp;
-  sec_dat->size = max_decomp_size;
-  sec_dat->version = dat->version;
-
   return 0;
 }
 
@@ -806,7 +845,7 @@ read_sections_map (Bit_Chain *dat, int64_t size_comp, int64_t size_uncomp,
 
   while (ptr < ptr_end)
     {
-      section = (r2007_section *)malloc (sizeof (r2007_section));
+      section = (r2007_section *)calloc (1, sizeof (r2007_section));
       if (!section)
         {
           LOG_ERROR ("Out of memory");
@@ -827,18 +866,24 @@ read_sections_map (Bit_Chain *dat, int64_t size_comp, int64_t size_uncomp,
       LOG_TRACE ("  num pages:     %" PRIu64 "\n", section->num_pages);
 
       // debugging sanity
-#if 0
+#if 1
       /* compressed */
-      if (section->data_size > 2 * dat->size)
-        return DWG_ERR_VALUEOUTOFBOUNDS;
-      assert(section->data_size < dat->size + 0x100000);
-      assert(section->max_size  < dat->size + 0x100000);
-      assert(section->name_length < dat->size);
-      assert(section->num_pages < DBG_MAX_COUNT);
+      if (section->data_size > 10 * dat->size
+          || section->name_length >= (int64_t)dat->size)
+        {
+          LOG_ERROR ("Invalid System Section");
+          free (section);
+          free (data);
+          sections_destroy (sections); // the root
+          return NULL;
+        }
+        // assert(section->data_size < dat->size + 0x100000);
+        // assert(section->max_size  < dat->size + 0x100000);
+        // assert(section->num_pages < DBG_MAX_COUNT);
 #endif
-      section->next = NULL;
-      section->pages = NULL;
-      section->name = NULL;
+      // section->next = NULL;
+      // section->pages = NULL;
+      // section->name = NULL;
 
       if (!sections)
         {
@@ -869,8 +914,8 @@ read_sections_map (Bit_Chain *dat, int64_t size_comp, int64_t size_uncomp,
       if (section->num_pages <= 0)
         continue;
 
-      section->pages = (r2007_section_page **)malloc (
-          (size_t)section->num_pages * sizeof (r2007_section_page *));
+      section->pages = (r2007_section_page **)calloc (
+          (size_t)section->num_pages, sizeof (r2007_section_page *));
       if (!section->pages)
         {
           LOG_ERROR ("Out of memory");
@@ -884,7 +929,7 @@ read_sections_map (Bit_Chain *dat, int64_t size_comp, int64_t size_uncomp,
       for (i = 0; i < section->num_pages; i++)
         {
           section->pages[i]
-              = (r2007_section_page *)malloc (sizeof (r2007_section_page));
+              = (r2007_section_page *)calloc (1, sizeof (r2007_section_page));
           if (!section->pages[i])
             {
               LOG_ERROR ("Out of memory");
@@ -907,6 +952,14 @@ read_sections_map (Bit_Chain *dat, int64_t size_comp, int64_t size_uncomp,
           LOG_HANDLE (" checksum: %016" PRIx64, section->pages[i]->checksum);
           LOG_HANDLE (" crc64: %016" PRIx64 "\n", section->pages[i]->crc);
           // debugging sanity
+          if (section->pages[i]->size >= DBG_MAX_SIZE
+              || section->pages[i]->uncomp_size >= DBG_MAX_SIZE
+              || section->pages[i]->comp_size >= DBG_MAX_SIZE)
+            {
+              LOG_ERROR ("Invalid section->pages[%d].*size", i);
+              section->num_pages = i; // skip this last section
+              return sections;
+            }
           assert (section->pages[i]->size < DBG_MAX_SIZE);
           assert (section->pages[i]->uncomp_size < DBG_MAX_SIZE);
           assert (section->pages[i]->comp_size < DBG_MAX_SIZE);
@@ -1067,6 +1120,8 @@ read_file_header (Bit_Chain *restrict dat,
   memset (file_header, 0, sizeof (r2007_file_header));
   bit_read_fixed (dat, data, 0x3d8);
   pedata = decode_rs (data, 3, 239);
+  if (!pedata)
+    return DWG_ERR_OUTOFMEM;
 
   // Note: This is unportable to big-endian
   seqence_crc = *((uint64_t *)pedata);
@@ -1118,15 +1173,64 @@ read_file_header (Bit_Chain *restrict dat,
       VALID_SIZE (file_header->pages_map_size_uncomp);
       VALID_COUNT (file_header->pages_maxid);
       VALID_COUNT (file_header->pages_amount);
-      VALID_COUNT (file_header->sections_amount);
+      VALID_COUNT (file_header->num_sections);
     }
 
   free (pedata);
   return error;
 }
 
+/* Return the latest dat position for all three independent streams
+   data, handle and string.
+*/
+unsigned long
+obj_stream_position (Bit_Chain *restrict dat, Bit_Chain *restrict hdl_dat,
+                     Bit_Chain *restrict str_dat)
+{
+  unsigned long p1 = bit_position (dat);
+  SINCE (R_2007)
+  {
+    /* all 3 now relative to obj */
+    unsigned long p2 = bit_position (hdl_dat);
+    unsigned long p3 = bit_position (str_dat);
+
+    if (p2 > p1)
+      return p3 > p2 ? p3 : p2;
+    else
+      return p3 > p1 ? p3 : p1;
+  }
+  else { return p1; }
+}
+
 int
-obj_string_stream (Bit_Chain *dat, Dwg_Object *restrict obj, Bit_Chain *str)
+obj_handle_stream (Bit_Chain *restrict dat, Dwg_Object *restrict obj,
+                   Bit_Chain *restrict hdl_dat)
+{
+  long unsigned int bit8 = obj->bitsize / 8;
+  long unsigned int pos = dat->byte;
+  // The handle stream offset, i.e. end of the object, right after
+  // the has_strings bit.
+  obj->hdlpos = obj->bitsize; // relative to dat
+  // restrict it to 0-end
+  hdl_dat->byte = bit8;
+  hdl_dat->bit = obj->bitsize % 8;
+  // bit_reset_chain (hdl_dat); //but keep the same start
+  if (!obj->handlestream_size)
+    {
+      obj->handlestream_size = (obj->size * 8) - obj->bitsize;
+      LOG_TRACE (" Hdlsize: %lu,", obj->handlestream_size);
+    }
+  hdl_dat->size = obj->size;
+  pos = (pos * 8) + obj->bitsize + obj->handlestream_size;
+  LOG_HANDLE (" hdl_dat: @%lu.%u - @%lu.%lu (%lu)", bit8, hdl_dat->bit,
+              pos / 8, pos % 8, hdl_dat->size);
+  LOG_TRACE ("\n")
+  return 0;
+}
+
+int
+obj_string_stream (Bit_Chain *restrict dat, Dwg_Object *restrict obj,
+                   Bit_Chain *restrict str)
 {
   BITCODE_RL start = obj->bitsize - 1; // in bits
   BITCODE_RL data_size = 0;            // in byte
@@ -1136,27 +1240,40 @@ obj_string_stream (Bit_Chain *dat, Dwg_Object *restrict obj, Bit_Chain *str)
   old_byte = str->byte;
 
   str->chain += str->byte;
+  // obj->strpos = str->byte * 8 + str->bit;
+
   str->byte = 0;
   str->bit = 0;
   str->size = (obj->bitsize / 8) + ((obj->bitsize % 8) ? 1 : 0);
   bit_advance_position (str, start - 8);
+
+  if (obj->supertype == DWG_SUPERTYPE_UNKNOWN)
+    {
+      str->size = 0;
+      // bit_reset_chain (str);
+      return 0;
+    }
   if (str->byte >= old_size - old_byte)
     {
       LOG_WARN ("obj_string_stream overflow");
       return DWG_ERR_VALUEOUTOFBOUNDS;
     }
-  LOG_TRACE (" obj string stream +%u: @%lu.%u (%lu)", start, str->byte,
-             str->bit & 7, bit_position (str));
+  LOG_HANDLE (" obj string stream +%u: @%lu.%u (%lu)", start, str->byte,
+              str->bit & 7, bit_position (str));
   obj->has_strings = bit_read_B (str);
   LOG_TRACE (" has_strings: %d\n", (int)obj->has_strings);
   if (!obj->has_strings)
-    return 0;
+    {
+      str->size = 0;
+      // bit_reset_chain (str);
+      return 0;
+    }
 
   bit_advance_position (str, -1); //-17
   str->byte -= 2;
   LOG_HANDLE (" @%lu.%u", str->byte, str->bit & 7);
   data_size = (BITCODE_RL)bit_read_RS (str);
-  LOG_HANDLE (" data_size: %u/0x%x\n", data_size, data_size);
+  LOG_HANDLE (" data_size: %u/0x%x", data_size, data_size);
 
   if (data_size & 0x8000)
     {
@@ -1169,19 +1286,24 @@ obj_string_stream (Bit_Chain *dat, Dwg_Object *restrict obj, Bit_Chain *str)
       LOG_HANDLE (" data_size: %u/0x%x\n", data_size, data_size);
       // LOG_TRACE("  -33: @%lu\n", str->byte);
     }
+  else
+    LOG_HANDLE ("\n");
   str->byte -= 2;
   if (data_size > obj->bitsize)
     {
-      LOG_TRACE ("Invalid string stream data_size: @%lu.%u\n", str->byte,
-                 str->bit & 7);
+      LOG_WARN ("Invalid string stream data_size: @%lu.%u\n", str->byte,
+                str->bit & 7);
       obj->has_strings = 0;
+      bit_reset_chain (str);
       return DWG_ERR_NOTYETSUPPORTED; // a very low severity error
     }
   obj->stringstream_size = data_size;
   bit_advance_position (str, -(int)data_size);
+  // bit_reset_chain (str);
   // LOG_TRACE(" %d: @%lu.%u (%lu)\n", -(int)data_size - 16, str->byte,
   // str->bit & 7,
   //          bit_position(str));
+  // obj->strpos = obj->bitsize_pos + obj->bitsize - obj->stringstream_size;
   return 0;
 }
 
@@ -1282,7 +1404,7 @@ read_2007_section_classes (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
       c = bit_read_B (&sec_dat); // 1
       LOG_HANDLE ("c: " FORMAT_B " [B]\n", c);
 
-      dwg->layout_number = 0;
+      dwg->layout_type = 0;
       dwg->num_classes = max_num - 499;
       if (max_num < 500 || max_num > 5000)
         {
@@ -1338,8 +1460,9 @@ read_2007_section_classes (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
 
           dwg->dwg_class[i].dxfname
               = bit_convert_TU (dwg->dwg_class[i].dxfname_u);
-          if (strcmp (dwg->dwg_class[i].dxfname, "LAYOUT") == 0)
-            dwg->layout_number = dwg->dwg_class[i].number;
+          if (dwg->dwg_class[i].dxfname
+              && strEQc (dwg->dwg_class[i].dxfname, "LAYOUT"))
+            dwg->layout_type = dwg->dwg_class[i].number;
         }
     }
   else
@@ -1467,22 +1590,21 @@ read_2007_section_handles (Bit_Chain *dat, Bit_Chain *hdl,
           return DWG_ERR_VALUEOUTOFBOUNDS;
         }
 
-      // last_handle = 0;
       last_offset = 0;
       while (hdl_dat.byte - startpos < section_size)
         {
           int added;
-          BITCODE_UMC handle;
+          BITCODE_UMC handleoff;
           BITCODE_MC offset;
 
           oldpos = hdl_dat.byte;
-          handle = bit_read_UMC (&hdl_dat);
+          handleoff = bit_read_UMC (&hdl_dat);
           offset = bit_read_MC (&hdl_dat);
-          // last_handle += handle;
           last_offset += offset;
-          LOG_TRACE ("\nNext object: %lu\t", (unsigned long)dwg->num_objects)
-          LOG_TRACE ("Handle: %lX\tOffset: " FORMAT_MC " @%lu\n", handle,
-                     offset, last_offset)
+          LOG_TRACE ("\nNext object: %lu ", (unsigned long)dwg->num_objects)
+          LOG_TRACE ("Handleoff: " FORMAT_MC " [UMC] "
+                     "Offset: " FORMAT_MC " [MC] @%lu\n",
+                     handleoff, offset, last_offset)
 
           if (hdl_dat.byte == oldpos)
             break;
@@ -1508,9 +1630,9 @@ read_2007_section_handles (Bit_Chain *dat, Bit_Chain *hdl,
         }
       else
         {
-          LOG_WARN (
-              "Handles section page CRC: %04X vs calc. %04X from %lx-%lx\n",
-              crc2, crc1, startpos, hdl_dat.byte - 2);
+          LOG_WARN ("Handles section page CRC mismatch: %04X vs calc. %04X "
+                    "from %lx-%lx\n",
+                    crc2, crc1, startpos, hdl_dat.byte - 2);
           error |= DWG_ERR_WRONGCRC;
         }
 #endif
@@ -1527,6 +1649,92 @@ read_2007_section_handles (Bit_Chain *dat, Bit_Chain *hdl,
   return error;
 }
 
+static int
+read_2007_section_summary (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
+                           r2007_section *restrict sections_map,
+                           r2007_page *restrict pages_map)
+{
+  static Bit_Chain old_dat, obj_dat = { 0 };
+  Bit_Chain *str_dat;
+  struct Dwg_SummaryInfo *_obj = &dwg->summaryinfo;
+  Dwg_Object *obj = NULL;
+  int error;
+  BITCODE_RL rcount1 = 0, rcount2 = 0;
+
+  old_dat = *dat;
+  error = read_data_section (&obj_dat, dat, sections_map, pages_map,
+                             SECTION_SUMMARYINFO);
+  if (error >= DWG_ERR_CRITICAL)
+    {
+      LOG_ERROR ("Failed to read SummaryInfo section");
+      if (obj_dat.chain)
+        free (obj_dat.chain);
+      return error;
+    }
+
+  if (dwg->header.summaryinfo_address != (BITCODE_RL)dat->byte)
+    LOG_WARN ("summaryinfo_address mismatch: " FORMAT_RL " != %lu",
+              dwg->header.summaryinfo_address, dat->byte);
+  LOG_TRACE ("\nSummaryInfo\n-------------------\n")
+  str_dat = dat = &obj_dat; // restrict in size
+
+  // clang-format off
+  #include "summaryinfo.spec"
+  // clang-format on
+
+  if (obj_dat.chain)
+    free (obj_dat.chain);
+  *dat = old_dat; // unrestrict
+  return error;
+}
+
+static int
+read_2007_section_preview (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
+                           r2007_section *restrict sections_map,
+                           r2007_page *restrict pages_map)
+{
+  static Bit_Chain sec_dat = { 0 };
+  int error;
+  BITCODE_RL size;
+  unsigned char *sentinel;
+
+  error = read_data_section (&sec_dat, dat, sections_map, pages_map,
+                             SECTION_PREVIEW);
+  if (error >= DWG_ERR_CRITICAL)
+    {
+      LOG_ERROR ("Failed to read uncompressed %s section", "Preview");
+      if (sec_dat.chain)
+        free (sec_dat.chain);
+      return error;
+    }
+
+  if (dwg->header.thumbnail_address != (BITCODE_RL)dat->byte)
+    LOG_WARN ("thumbnail_address mismatch: " FORMAT_RL " != %lu",
+              dwg->header.thumbnail_address, dat->byte);
+  LOG_TRACE ("\nPreview\n-------------------\n")
+
+  sentinel = dwg_sentinel (DWG_SENTINEL_THUMBNAIL_BEGIN);
+  if (memcmp (sentinel, sec_dat.chain, 16))
+    {
+      LOG_WARN ("thumbnail sentinel mismatch");
+      return error |= DWG_ERR_WRONGCRC;
+    }
+
+  dwg->thumbnail.size = sec_dat.size - 32; // 2x sentinel
+  dwg->thumbnail.chain = sec_dat.chain;
+  dwg->thumbnail.byte = 16; // sentinel
+
+  dwg_bmp (dwg, &size);
+  if (abs ((int)((long)size - (long)dwg->thumbnail.size))
+      > 200) // various headers
+    LOG_WARN ("thumbnail.size mismatch: %lu != " FORMAT_RL,
+              dwg->thumbnail.size, size);
+
+  dat->byte += dwg->thumbnail.size;
+
+  return error;
+}
+
 /* exported */
 void
 read_r2007_init (Dwg_Data *dwg)
@@ -1541,7 +1749,7 @@ read_r2007_meta_data (Bit_Chain *dat, Bit_Chain *hdl_dat,
 {
   r2007_file_header file_header;
   r2007_page *restrict pages_map, *restrict page;
-  r2007_section *restrict sections_map;
+  r2007_section *restrict sections_map = NULL;
   int error;
 #ifdef USE_TRACING
   char *probe;
@@ -1566,7 +1774,7 @@ read_r2007_meta_data (Bit_Chain *dat, Bit_Chain *hdl_dat,
                               file_header.pages_map_size_uncomp,
                               file_header.pages_map_correction);
   if (!pages_map)
-    return DWG_ERR_PAGENOTFOUND;
+    return DWG_ERR_PAGENOTFOUND; // Error already logged
 
   // Sections Map
   page = get_page (pages_map, file_header.sections_map_id);
@@ -1574,8 +1782,8 @@ read_r2007_meta_data (Bit_Chain *dat, Bit_Chain *hdl_dat,
     {
       LOG_ERROR ("Failed to find sections page map %d",
                  (int)file_header.sections_map_id);
-      pages_destroy (pages_map);
-      return DWG_ERR_SECTIONNOTFOUND;
+      error |= DWG_ERR_SECTIONNOTFOUND;
+      goto error;
     }
   dat->byte = page->offset;
   if ((unsigned long)file_header.sections_map_size_comp
@@ -1585,21 +1793,37 @@ read_r2007_meta_data (Bit_Chain *dat, Bit_Chain *hdl_dat,
                  __FUNCTION__,
                  (unsigned long)file_header.sections_map_size_comp,
                  dat->size - dat->byte)
-      return DWG_ERR_VALUEOUTOFBOUNDS;
+      error |= DWG_ERR_VALUEOUTOFBOUNDS;
+      goto error;
     }
   sections_map = read_sections_map (dat, file_header.sections_map_size_comp,
                                     file_header.sections_map_size_uncomp,
                                     file_header.sections_map_correction);
+  if (!sections_map)
+    goto error;
 
-  error = read_2007_section_classes (dat, dwg, sections_map, pages_map);
   error
-      += read_2007_section_header (dat, hdl_dat, dwg, sections_map, pages_map);
-  error += read_2007_section_handles (dat, hdl_dat, dwg, sections_map,
+      = read_2007_section_header (dat, hdl_dat, dwg, sections_map, pages_map);
+  if (dwg->header.summaryinfo_address)
+    error |= read_2007_section_summary (dat, dwg, sections_map, pages_map);
+  error |= read_2007_section_classes (dat, dwg, sections_map, pages_map);
+  error |= read_2007_section_handles (dat, hdl_dat, dwg, sections_map,
                                       pages_map);
-  // read_2007_blocks(dat, hdl_dat, dwg, sections_map, pages_map);
+  if (dwg->header.thumbnail_address)
+    error |= read_2007_section_preview (dat, dwg, sections_map, pages_map);
+  // if (dwg->header.vbaproj_address)
+  //  error |= read_2007_section_vbaproject (dat, dwg, sections_map,
+  //  pages_map);
+  // error |= read_2007_section_appinfo (dat, dwg, sections_map, pages_map);
+  // error |= read_2007_section_filedeplist (dat, dwg, sections_map,
+  // pages_map); error |= read_2007_section_security (dat, dwg, sections_map,
+  // pages_map); error |= read_2007_section_revhistory (dat, dwg, sections_map,
+  // pages_map);
+  // read_2007_blocks (dat, hdl_dat, dwg, sections_map, pages_map);
 
+error:
   pages_destroy (pages_map);
-  if (page)
+  if (sections_map)
     sections_destroy (sections_map);
 
   return error;
